package cz.slabi.david.animaletto;

import java.net.*;
import java.util.*;
import java.io.*;

/**
 * The network client used to communicate with the server. The brain of whole
 * application. Keeps the game state, reads and sends messages, contains
 * method to handle non-visual card actions.<br />
 * <br />
 * When a message or a set of messages should be send to the server, it's
 * kept in {@code littleQueue}. If a response doesn't come soon, this
 * queue is used to re-send the messages.<br />
 * <br />
 * During {@code DECIDE} game state, which is set upon first connection attempt,
 * client decides according to information acquired from server whether
 * a new session will be established or if there is one already to
 * connect to. The game state is switched to {@code CREATE} or {@code CONNECT}. After
 * creating or connecting {@code SETUP} game state is set. During this state
 * player-specific variables are set. Then comes the {@code WAIT} game state.
 * In this state application stays until all players are ready to start
 * and information about being prepared is send to server during this state.
 * If everyone is ready, {@code PREPARE} game state is set. 
 * Number of beginning player and random seed to prepare same 
 * decks are acquired. Since this state automatic check of players'
 * names is performed to notice someone has disconnected. 
 * If it happens, game isinterrupted. If the local player
 * is on turn, switches to {@code PLAY} game state, else switches to {@code WATCH}.
 * During these states game is performed; information about picked
 * cards or grabbed packets are sent and recieved. If a player finishes
 * his turn, increments {@code turn} variable on server and sets next
 * player's variables so that no synchronization problems may occur
 * here. {@code FINISH} state is the one where jokers are put at chosen
 * packets. During {@code SLEEP} state no action is performed; it's 
 * the state when players watch the score table.<br />
 *
 * @author DavidSlaby
 */

public class Client{
	
    /**
     * Basic states of the application.
     */
	public enum GameState { NONE, DECIDE, CONNECT, CREATE, SETUP, WAIT, PREPARE, 
		PLAY, WATCH, FINISH, SLEEP, QUIT }
    
    /**
     * Actions performed with cards during the game.
	 * Used during PLAY and WATCH game state.
     */
	public enum CardAction { NONE, PICKCARD, GETPACKET, PUTCARD }
	
	private static String[] usedVars = { "ready", "packet", "decid", "done", "joker" };
	private static String[] usedVarsVals = { "false", "-1", "-2", "false", "_" };
	
    private static final int BUFLEN = 256;

	private static final String CONFIG_FILE = "data/animaletto.conf";

	private static final int SESSION_VARIABLES = 64;
	private static final int SESSION_TIMEOUT = 180;
	private static final int SESSION_MAXPLAYERS = 6;
    
    public static final int GAMES_COUNT = 4;

	public DatagramSocket socket;
	
	public String name;
	public String session;
	
	public InetSocketAddress server;
	
    private LinkedList<String> inQueue, outQueue;
    private LinkedList<String> littleQueue;
	
	private Sender sender;
	private Reader reader;
	private Reciever reciever;

    private int checkInterval = 40;

    private int latency;
	private int playerCheckLatency = 0;
    private final int maxLatency = 2000;
	private final int maxPlayerCheckLatency = 2000;

	public GameState gameState;
	public CardAction cardAction;
	
	private boolean lateDanger;
    
    public boolean isReady;
	
	private Random randomizer;
	
	public int myPosition;
    public int playersCount;
	public String[] playerNames;
	
	public Deck deck;
	public Card[][] packets;
	
	public Card picked;
	
	public Card[][] owned;
	public int[] ownedCount;
	
	public int actualPlayer;
	
	public boolean[] playersFinished;
    
    public int turn;
   	public int game;
    public int round;
	public boolean lastRound;
    
    public int[][] score;

    /**
     * Thread which recieves messages, filters them and puts them into the 
     * in-queue.
     */
	private class Reciever extends Thread{
		
        /**
         * The thread method which is recieving messages.
         * 
         */
		@Override
		public void run(){
			while (true){
				while (!socket.isClosed()){

					byte[] buf = new byte[BUFLEN];

					DatagramPacket in = new DatagramPacket(buf, buf.length);
					try{
						if (!socket.isClosed()) socket.receive(in);
					}catch(Exception e){
						continue;
					}

					String[] split = new String(in.getData()).split("\0", 2);

					String received = split[0];

					if (in.getSocketAddress().equals(server))
						if (socket.isClosed()){
							inQueue.clear();
							return;
						} else 
							inQueue.offer(received);
					else
						System.out.println("enemy message!");

					try{
						sleep(40);
					}catch(InterruptedException e){
					}
				}
				
				try{
					sleep(40);
				}catch(InterruptedException e){
					
				}

			}
		}
		
	}
	
    /**
     * Thread which reads recieved messages and does actions according to its
     * contents and the game state. Contains most of the logics of the application.
     * 
     */
	private class Reader extends Thread{
		
		boolean maybeStart = false;
		String msg;
		String[] elements;
		
        /**
         * Works with the messages from the in-queue and controls the game; also
         * checks the latency and resends lost packets.
		 * 
		 * Takes messages from the in-queue and does an action according to its
		 * content and actual game state. If there it no message in the queue, 
		 * performs one "empty pass" to check whether there are network latency
		 * problems.
		 * 
         */
		public void doPending(){
			boolean emptyPass = true;
			boolean tryAgain = false;
			
			while ((!inQueue.isEmpty()) || emptyPass){
				
				emptyPass = false;

                playerCheckLatency += checkInterval;
                
                setHint();

				if (!inQueue.isEmpty()){
					msg = inQueue.poll();
					elements = msg.split(" +");
					latency = 0;
				}else{
					msg = "";
					elements = new String[2];
					elements[0] = " "; elements[1] = " ";
					latency += checkInterval;
				}
				
				switch (gameState){
					
					case DECIDE:
						
						if (elements[0].equals("s")){
							
							boolean sessionExists = false;
							
							for (int i = 2; i < elements.length; i++)
								if (session.equals(elements[i])){
									sessionExists = true; break;
								}
							
							littleQueue.clear();
							if (sessionExists){
								gameState = GameState.CONNECT;
								littleQueue.offer("c "+name+" "+session);
							}else{
								gameState = GameState.CREATE;
								littleQueue.offer("n "+name+" "+session
                                        +" "+String.valueOf(SESSION_VARIABLES)
                                        +" "+String.valueOf(SESSION_TIMEOUT)
                                        +" "+String.valueOf(SESSION_MAXPLAYERS));
							}
							mailBox();
						}
						break;
						
					case CONNECT:
						
						if ((elements[0].equals("i")) && (elements[1].equals("c"))){
							Main.mainWindow.logRecord("Connected to session.");
							gameState = GameState.SETUP;
							littleQueue.clear();
							littleQueue.offer("p "+initVars());
							mailBox();
							Main.mainWindow.gotConnected();
						}else if ((elements[0].equals("e")) && (elements[1].equals("c"))){
							elements = msg.split(" +", 3);
							Main.mainWindow.logRecord("Error: " + elements[2].substring(1));
							gameState = GameState.NONE;
							Main.mainWindow.disConnect();
						}
						break;
						
					case CREATE:
						
						if ((elements[0].equals("i")) && (elements[1].equals("n"))){
							Main.mainWindow.logRecord("Session established.");
							gameState = GameState.SETUP;
							littleQueue.clear();
							littleQueue.offer("p "+initVars()+" turn -1");
							mailBox();
							Main.mainWindow.gotConnected();
						} else if ((elements[0].equals("e")) && (elements[1].equals("n"))){
							elements = msg.split(" +", 3);
							Main.mainWindow.logRecord("Error: " + elements[2].substring(1));
							gameState = GameState.NONE;
							Main.mainWindow.disConnect();
						}
						break;
						
					case SETUP:
						
						if ((elements[0].equals("i")) && (elements[1].equals("p"))){
                            gameState = GameState.WAIT;
                            lateDanger = true;
                            waitLittleQueueCreate();
							mailBox();
							checkInterval = 250;
						}else if ((elements[0].equals("e")) && (elements[1].equals("p"))){
							tryAgain = true;
						}
						break;
						
					case WAIT:
						
						if (elements[0].equals("w")){
							int playerCnt = Integer.parseInt(elements[4]);
							String[] newPlayers = new String[playerCnt];
							System.arraycopy(elements, 5, newPlayers, 0, playerCnt);
							if (!Arrays.equals(playerNames, newPlayers)){
								maybeStart = false;
								playerNames = newPlayers;
								StringBuilder playersMsg = new StringBuilder("");
								for(String pl: playerNames)
									playersMsg.append(pl).append(" ");
								
                                Main.mainWindow.logRecord("Players: "+playersMsg.toString());

                                waitLittleQueueCreate();

                                maybeStart = false;
                                
							}else if (maybeStart){
                                
									checkInterval = 250;
									gameState = GameState.PREPARE;
									Main.mainWindow.prepareGame();
									for (int i=0; i<playerNames.length; i++)
										if (playerNames[i].equals(name)) myPosition = i;
									
									littleQueue.clear();

									if (myPosition == 0){
										turn = (int)(Math.random() * playerNames.length);
										int randomSeed = (int)(Math.random() * Integer.MAX_VALUE);
										littleQueue.offer("p randomseed "+String.valueOf(randomSeed)
											+" turn "+String.valueOf(turn));
									}
									
									littleQueue.offer("g randomseed turn");
							}
							mailBox();
							
						}else if (lateDanger &&
                            (elements[0].equals("g")) && (elements[1].equals("turn"))){
                            if (elements[2].equals("-1")){
                                lateDanger = false;
                                waitLittleQueueCreate();
                            }else{
                                Main.mainWindow.logRecord("Game already started.");
                                Main.mainWindow.disConnect();
                            }
						
                        }else if ((elements[0].equals("g")) && (playerNames != null)){
							int readyCount = 0;
							for (int i = 1; i < elements.length - 1; i += 2){
								
								if ((elements[i].equals(name + "_ready")) &&
									((elements[i+1].equals("true")) != isReady)){
									littleQueue.offer("p " + name + "_ready " + (isReady ? "true" : "false"));
								}
								
								if ((elements[i].length() > 6) && 
                                        (elements[i].substring(elements[i].length() - 6).equals("_ready")) 
									&& (elements[i+1].equals("true"))){
									String plName = elements[i].substring(0, elements[i].length() - 6);
									for (String pl: playerNames){
										if (plName.equals(pl)){
                                            readyCount++;
                                            break;
                                        }
									}
								}
							}
                            
							if ((readyCount == playerNames.length) && (readyCount > 1) && !lateDanger){
								maybeStart = true;
							}
						}
						break;
						
					case PREPARE:
						
						if ((elements[0].equals("g")) && (elements[1].equals("randomseed"))
                            && (elements.length > 4) && (elements[3].equals("turn"))){

                            game = 0;
							turn = Integer.parseInt(elements[4]);
                            playersCount = playerNames.length;
                            
                            if (turn > playersCount) { mailBox(); break;}
                            
							actualPlayer = turn;
                            score = new int[playersCount][GAMES_COUNT+1];
							
							int randomSeed = Integer.parseInt(elements[2]);
                            randomizer = new Random(randomSeed);
                            
							picked = null;
							
                            nextGame();
						
						}else if ((elements[0].equals("e")) && (elements[1].equals("g"))){
							tryAgain = true;
						}
						break;
						
					case PLAY:
						
						if ((cardAction == CardAction.PUTCARD) ||
							(cardAction == CardAction.GETPACKET)){
							if ((elements[0].equals("g")) && (elements[1].equals(name+"_decid"))){
								if (elements[6].equals("true")){
									if (Integer.parseInt(elements[2]) >= 0)
										playersFinished[actualPlayer] = true;
									turnIncrement();
								}else{
									mailBox();
								}
							}
						}
						break;
						
					case WATCH:
						
						if ((elements[0].equals("g")) &&
							(elements[1].equals(playerNames[actualPlayer]+"_decid"))){
							
							if (Integer.parseInt(elements[8]) == turn){

								int decidePacket = Integer.parseInt(elements[2]);
								if (decidePacket == -1){
									if (picked == null) picked = deck.removeFirst();
									if (picked.type == Card.CardType.LAST){
										lastRound = true;
										picked = deck.removeFirst();
									}
									int whichPacket = Integer.parseInt(elements[4]);
									if (whichPacket >= 0){
										if (picked != null){
											for (int i = 0; i < 3; i++){
												if (packets[whichPacket][i] == null){
													packets[whichPacket][i] = picked;
													picked = null;
												}
											}
										}
									}
									Main.mainWindow.tableRefresh(false);
								}else if (decidePacket >= 0){
									int j = ownedCount[actualPlayer];
									for (int i = 0; i < packets[decidePacket].length; i++){
										if (packets[decidePacket][i] != null){
											owned[actualPlayer][j+i] = packets[decidePacket][i];
											ownedCount[actualPlayer] = j+i+1;
										}else break;
									}
									packets[decidePacket] = null;
									Main.mainWindow.tableRefresh(false);
								}

								if (elements[6].equals("true")){
									if (Integer.parseInt(elements[2]) >= 0)
										playersFinished[actualPlayer] = true;
									turnIncrement();
									
								}else mailBox();
								
							}else mailBox();
							
						}
						break;
						
					case FINISH:
						if (elements[0].equals("g")){

							boolean everyoneFinished = true;
							
							for (int i = 1; i < elements.length; i += 2){
								
								if ((elements[i].equals(name + "_joker")) &&
									(!(elements[i+1].equals(jokersToString())))){
									
									littleQueue.clear();
									littleQueue.offer("p "+name+"_joker "+jokersToString());

									StringBuilder jokerVars = new StringBuilder("");
									for(String pl: playerNames) jokerVars.append(pl).append("_joker ");
									littleQueue.offer("g "+jokerVars.toString());
								}
								
								if ((elements[i].length() >= 6) 
									&& (elements[i].substring(elements[i].length() - 6).equals("_joker"))){
									String plName = elements[i].substring(0, elements[i].length() - 6);
									String[] jokers = elements[i+1].split("_");
									
									if (jokers.length > 0){
										int jkr = 1;
										for (int j = 0; j < playersCount; j++)
											if (plName.equals(playerNames[j])){
												for (int ii = 0; ii < ownedCount[j]; ii++)
													if (owned[j][ii].type == Card.CardType.JOKER){
														int newColor = Integer.parseInt(jokers[jkr]);
														if (newColor == -1) everyoneFinished = false;
														if (j != myPosition) owned[j][ii].color = newColor;
														jkr++;
													}
													if (jokers[jokers.length - 1].equals("-1"))
														everyoneFinished = false;
											}
									}
                                    
								}
							}

							if (everyoneFinished){
								gameState = gameState.SLEEP;
								littleQueue.clear();
                                countScore();
								new ScoreDialog(new java.awt.Frame(), false);
							}

							Main.mainWindow.tableRefresh(false);
							mailBox();
						}
						break;
						
				} /* switch end */
				
                if ((gameState.ordinal() > GameState.SETUP.ordinal())
                    && (gameState.ordinal() < GameState.QUIT.ordinal())){
                    if (elements[0].equals("a")){
                        elements = msg.split(" +", 3);
                        Main.mainWindow.chatRecord(elements[1], elements[2].substring(1));
                    }
                    if (gameState.ordinal() > GameState.PREPARE.ordinal()){
                        if (elements[0].equals("w")){
                            
							int playerCnt = Integer.parseInt(elements[4]);
							String[] newPlayers = new String[playerCnt];
							System.arraycopy(elements, 5, newPlayers, 0, playerCnt);
							if (!Arrays.equals(playerNames, newPlayers)){
								finishedGames(false);
                            }
						}
                        if (playerCheckLatency > maxPlayerCheckLatency){
                            playerCheckLatency = 0;
                            outQueue.offer("w");
                        }
                    }
                }
                

				if ((latency > maxLatency) || tryAgain){
					latency = 0;
					mailBox();
				}
				
			}
		}
		
        /**
         * Waits for a small period of time and then calls doPending() method.
         */
		@Override
		public void run(){
			while (true){
				doPending();
				try{
					sleep(checkInterval);
				}catch(InterruptedException e){
				}
			}
		}
	}
    
	/**
     * Thread which gets messages from out-queue, formats them and sends them
     * to the server.
     */
	private class Sender extends Thread{
		
        /**
         * Works with the messages from the out-queue and sends them to
         * the server.
         */
		public void doPending(){
			
			while (!outQueue.isEmpty()){

				byte[] buf;

				buf = (outQueue.poll() + "\0").getBytes();
				int length = (buf.length > BUFLEN) ? BUFLEN : buf.length;
				buf[length - 1] = '\0';

				DatagramPacket out;
				try{
					out = new DatagramPacket(buf, length, server);
				}catch (Exception e){
					outQueue.clear();
					continue;
				}

				if (socket.isClosed()){
					outQueue.clear();
					return;
				}

				try{
					socket.send(out);
				}catch(Exception e){
					System.out.println("Malfromed out-packet");
					return;
				}
			}
		}
		
        
        /**
         * Waits for a small period of time and then calls doPending() method.
         */
		@Override
		public void run(){
			while (true){
				doPending();
				try{
					sleep(40);
				}catch(InterruptedException e){
				}
			}
		}
		
	}
    
    /**
     * Sets a hint for the player to be displayed in the main window according
     * to the state of game.
     */ 
    private void setHint(){
        String temp = new String("");
        switch (gameState){
            case NONE: 
                temp = "Verify game and network settings and then connect.";
                break;
            case DECIDE: 
                temp = "Wait until you're connected.";
                break;
            case WAIT: 
                temp = "Set 'Ready' in the 'Game' menu once you wish to start.";
                break;
            case PREPARE: 
                temp = "Wait while the game is being prepared.";
                break;
            case PLAY: 
                switch (cardAction){
                    case NONE:
                        temp = "You're on turn: Pick a card or get a packet.";
                        break;
                    case PICKCARD:
                        temp = "You're on turn: Select a packet.";
                        break;
                }
                break;
            case WATCH:
                if (playersFinished[myPosition])
                    temp = playerNames[actualPlayer]+" is on turn. Wait until end of round.";
                else
                    temp = playerNames[actualPlayer]+" is on turn. Wait until you're on turn.";
                break;
            case FINISH: 
                temp = "Select packets to put chameleons on.";
                break;
            case SLEEP: 
                temp = "Watch the score.";
                break;
        }
        if (Main.mainWindow != null) Main.mainWindow.setHint(temp);
    }
    
    /**
     * Creates a queue of outgoing messages created during waiting for other
     * players.
     */
    private void waitLittleQueueCreate(){
        littleQueue.clear();
		littleQueue.offer("w");
		if (lateDanger) littleQueue.offer("g turn");
        if (playerNames != null){
            StringBuilder readyVars = new StringBuilder("");
            for(String pl: playerNames) readyVars.append(pl).append("_ready ");
        	littleQueue.offer("g "+readyVars.toString());
        }
        
    }
    
    /**
     * Sets user-specific variable names with initial values that will be send
     * to the server when connecting to it.
     * 
     * @return  a string containing user-specific variables with initial values
     */
	private String initVars(){
		StringBuffer temp = new StringBuffer("");
		for (int i = 0; i < usedVars.length; i++)
			temp.append(name+"_"+usedVars[i]+" "+usedVarsVals[i]+" ");
		return temp.toString();
	}

    /**
     * Sets user-specific variable names that will be send to the server for
     * deletion. Used whed disconnecting from server.
     * 
     * @return  a string containing user-specific variables
     */
	private String removeVars(){
		StringBuffer temp = new StringBuffer("");
		for (int i = 0; i < usedVars.length; i++)
			temp.append(name+"_"+usedVars[i]+" ");
		return temp.toString();
	}
	
    /**
     * Performed when a player finishes his turn. Also handles end of round and
     * end of game.
     */
	private void turnIncrement(){
		int donePlayer = actualPlayer;
		do{
			turn++;
			actualPlayer = turn % playersCount;
		}while ((playersFinished[actualPlayer])
			&& (actualPlayer != donePlayer));

		littleQueue.clear();
		
		if ((actualPlayer == donePlayer) && (playersFinished[actualPlayer])){
			round++;
			
			if (lastRound){
                game++;
                
				gameState = GameState.FINISH;
				Main.mainWindow.tableRefresh(false);
				littleQueue.offer("p "+name+"_joker "+jokersToString());
				
				StringBuilder jokerVars = new StringBuilder("");
				for(String pl: playerNames) jokerVars.append(pl).append("_joker ");
				littleQueue.offer("g "+jokerVars.toString());

				return;
			}
			
			packets = new Card[playersCount][3];
			for (int i = 0; i < playersCount; i++) playersFinished[i] = false;
			Main.mainWindow.tableRefresh(false);
		}
		
		if (actualPlayer == myPosition){
			cardAction = CardAction.NONE;
			gameState = GameState.PLAY;
		}else{
			gameState = GameState.WATCH;
			littleQueue.offer("g "+playerNames[actualPlayer]+"_decid "
				+playerNames[actualPlayer]+"_packet "
				+playerNames[actualPlayer]+"_done turn");
			mailBox();
		}
	}
	
    /**
     * Initializes required variables for the next game.
     */
	public void nextGame(){
        round = 1;
        lastRound = false;
        playersFinished = new boolean[playersCount];
        deck = new Deck(randomizer, playersCount);
        packets = new Card[playersCount][3];

        owned = new Card[playersCount][deck.size()];
        ownedCount = new int[playersCount];

        for (int i = 0; i < owned.length; i++) playersFinished[i] = false;

        for (int i = 0; i < playersCount; i++){
            ownedCount[i] = 1;
            owned[i][0] = new Card(Card.CardType.COLOR, i);
            playersFinished[i] = false;
        }

        littleQueue.clear();
        if (actualPlayer == myPosition){
            gameState = GameState.PLAY;
            cardAction = CardAction.NONE;
        }else{
            gameState = GameState.WATCH;
            littleQueue.offer("g "+playerNames[actualPlayer]+"_decid "
                +playerNames[actualPlayer]+"_packet "
                +playerNames[actualPlayer]+"_done turn");
        }
        
        checkInterval = 100;

        Main.mainWindow.tableRefresh(true);
        
        Main.mainWindow.logRecord(playerNames[actualPlayer]+" begins.");
        
		mailBox();
	}
    
    /**
     * Performed after all games were finished or when someone disconnected. 
     * Puts game to the state before players set to be ready.
     * 
     * @param alright   true when all games finished, false when someone disconnected
     */
    public void finishedGames(boolean alright){
        gameState = GameState.SETUP;
		reader.maybeStart = false;
		littleQueue.clear();
        littleQueue.offer("p "+initVars()+" turn -1");
        littleQueue.offer("f randomseed");
		mailBox();
        isReady = false;
		deck = null;
        Main.mainWindow.finishedGames(alright);
    }
	
    /**
     * Puts messages from the little queue to the main out-queue.
     */
	private void mailBox(){
		for (String s: littleQueue) outQueue.offer(s);
	}

    /**
     * Constructs a string containing coded values of jokers positions. Used
     * during the final part of each game.
     * 
     * @return  string containing coded values of jokers positions
     */
	private String jokersToString(){
		String s = "_";
		for (int i = 0; i < ownedCount[myPosition]; i++)
			if (owned[myPosition][i].type == Card.CardType.JOKER)
				s = s.concat(String.valueOf(owned[myPosition][i].color)+"_");
		return s;
	}

	/**
     * Chooses the action to be performed when a player is on turn.
     * 
     * @param whichPacket   equals -1 if picking card, non-negative if getting packet
     */
    public void doPlayCardAction(int whichPacket){
		switch (cardAction){
			case NONE:
				if (whichPacket < 0){
					boolean existsFreePacket = false;
					for (int i = 0; i < playersCount; i++){
						if ((packets[i] != null) && (packets[i][2] == null)) existsFreePacket = true;
					}
					if (existsFreePacket){
						cardAction = CardAction.PICKCARD;
						picked = deck.removeFirst();
						if (picked.type == Card.CardType.LAST){
							lastRound = true;
							picked = deck.removeFirst();
						}
						littleQueue.clear();
						littleQueue.offer("p "+name+"_decid -1 "+name+"_packet -1 "
							+name+"_done false turn "+String.valueOf(turn));
						mailBox();
					}
				}else{
					cardAction = CardAction.GETPACKET;
					int j = ownedCount[myPosition];
					for (int i = 0; i < packets[whichPacket].length; i++){
						if (packets[whichPacket][i] != null){
							owned[myPosition][j+i] = packets[whichPacket][i];
							ownedCount[myPosition] = j+i+1;
						}else break;
					}
					packets[whichPacket] = null;
					
					String nextPlayer = playerNames[(myPosition + 1) % playersCount];					
					
					littleQueue.clear();
					littleQueue.offer("p "
						+name+"_decid "+whichPacket+" "+name+"_packet -1 "+name+"_done true "
						+"turn "+String.valueOf(turn)+" "
						+nextPlayer+"_done false "+nextPlayer+"_decid -2 "+nextPlayer+"_packet -1 "
							+name+"_joker "+jokersToString());
					littleQueue.offer("g "+name+"_decid "+name+"_packet "+name+"_done turn");
					mailBox();
					
				}
				break;
			case PICKCARD:
				for (int i = 0; i < 3; i++){
					if (packets[whichPacket][i] == null){
						cardAction = CardAction.PUTCARD;
						packets[whichPacket][i] = picked;
						picked = null;
						
						String nextPlayer = playerNames[(myPosition + 1) % playersCount];
						littleQueue.clear();
						littleQueue.offer("p "
							+name+"_decid -1 "+name+"_packet "+whichPacket+" "+name+"_done true "
							+"turn "+String.valueOf(turn)+" "
							+nextPlayer+"_done false "+nextPlayer+"_decid -2 "+nextPlayer+"_packet -1");
						littleQueue.offer("g "+name+"_decid "+name+"_packet "+name+"_done turn");
						mailBox();
						break;
					}
				}
				break;
		}
		
	}
	
    /**
     * Puts joker to the selected owned packet.
     * 
     * @param whichPacket   equals the packet number, -1 if finishing without putting on packet
     */
	public void doFinishCardAction(int whichPacket){
		
        for (int i = 0; i < ownedCount[myPosition]; i++)
            if ((owned[myPosition][i].type == Card.CardType.JOKER) && (owned[myPosition][i].color == -1)){
                owned[myPosition][i].color = whichPacket;
                if (whichPacket >= 0) break;
            }
        
        littleQueue.clear();
        littleQueue.offer("p "+name+"_joker "+jokersToString());

        StringBuilder jokerVars = new StringBuilder("");
        for(String pl: playerNames) jokerVars.append(pl).append("_joker ");
        littleQueue.offer("g "+jokerVars.toString());
		
		mailBox();

	}
    
    /**
     * Prepares the score for ScoreDialog.
     */
    private void countScore(){
        for (int i = 0; i < playersCount; i++){
            
            int thisScore = 0;
            
            int bonuses = 0;
            int[] herdSize = new int[deck.colorCount + 1];
            
            for (int j = 0; j < ownedCount[i]; j++){
                if (owned[i][j].type == Card.CardType.BONUS)
                    bonuses++;
                else if (owned[i][j].color == -2)
                    herdSize[deck.colorCount]++;
                else if (owned[i][j].color >= 0)
                    herdSize[owned[i][j].color]++;
            }
            
            Arrays.sort(herdSize);
            for (int j = deck.colorCount; (j >= 0) && herdSize[j] > 0; j--){
                if (j >  deck.colorCount - 3)
                    thisScore += getScoreBySize(herdSize[j]);
                else
                    thisScore -= getScoreBySize(herdSize[j]);
            }
            thisScore += 2 * bonuses;
            
            score[i][game] = thisScore;
            
            thisScore = 0;
            for (int j = 1; j <= game; j++)
                thisScore += score[i][j];
            score[i][0] = thisScore;
            
        }
    }
    
    /**
     * Simply computes the value of packet according to its size.
     * 
     * @param size  number of cards in the owned packet
     * @return      value of the owned packet
     */
    private int getScoreBySize(int size){
        size = (size > 6) ? 6 : size;
        return (size + 1) * size / 2;
    }

    /**
     * Adds a chat message to the out-queue.
     * 
     * @param msg   message to be sent to other players
     */
	public void chat(String msg){
		outQueue.offer("a $"+msg);
	}
	
    /**
     * Sets the IP and Port of the server.
     * 
     * @param serverIP      IP address of the server
     * @param serverPort    port number of the server
     * @throws java.net.UnknownHostException    if IP address or port is invalid
     */
	public void setNetwork(String serverIP, int serverPort) throws UnknownHostException{
		server = new InetSocketAddress(InetAddress.getByName(serverIP), serverPort);
	}
	
    /**
     * Creates socket for communication with the server.
     * 
     * @throws java.net.SocketException     if socket creation failed
     */
	public void start() throws SocketException{
		socket = new DatagramSocket();

		latency = 0;
		playerNames = null;
		isReady = false;

		gameState = GameState.DECIDE;
		littleQueue.offer("s");
		mailBox();
	}
	
    /**
     * Sends disconnect message to the server and stops the network activity.
     */
	public void stop(){
		gameState = GameState.QUIT;
		outQueue.offer("f "+removeVars());
		outQueue.offer("disconnect");
		sender.doPending();
		socket.close();
		inQueue.clear();
		outQueue.clear();
		gameState = GameState.NONE;
		deck = null;
	}
	
	/**
     * Tests whether the socket might be created.
     * 
     * @throws java.net.SocketException     If the network settings are incorrect.
     */
    public void testSocket() throws SocketException{
		socket = new DatagramSocket();
		socket.close();
	}

	/**
	 * Reads network and game settings from the file.
	 */
	void loadSettings(){
		
		BufferedReader in = null;
		
		try{
			in = new BufferedReader(new FileReader(CONFIG_FILE));
			try{
				name = in.readLine();
				session = in.readLine();
				setNetwork(in.readLine(),
					Integer.valueOf(in.readLine()));
				in.close();
			}catch(IOException e){
				name = "someone";
				session = "animaletto";
				try{
					setNetwork("localhost", 9930);
				}catch(Exception eNet){
					System.err.println("Socket error");
				}
				System.err.println("Config file corrupted");
			}
		}catch (FileNotFoundException e){
			System.err.println("Config file not found!");
			System.exit(0);
		}
	}

	/**
	 * Saves network and game settings to the file.
	 */
	void saveSettings(){
		
		PrintWriter out = null;
		
		try{
			out = new PrintWriter(new FileWriter(CONFIG_FILE));
			out.println(name);
			out.println(session);
			out.println(server.getHostName());
			out.println(String.valueOf(server.getPort()));
			out.close();
		}catch(IOException e){
			System.err.println("Config file not writeable");
		}
		
	}

    /**
     * Tests network and initializes variables and threads.
     */
	public Client(){
		
		loadSettings();
		
		try{
			testSocket();
		}catch(SocketException e){
			System.out.println("Socket creation error");
		}
		
		inQueue = new LinkedList<String>();
		outQueue = new LinkedList<String>();
		littleQueue = new LinkedList<String>();
		
		gameState = GameState.NONE;
		cardAction = CardAction.NONE;
		
		reciever = new Reciever();
		reciever.start();

		reader = new Reader();
		reader.start();

		sender = new Sender();
		sender.start();
		
	}
}
